/* FAT Recover
   By Steven Lawrance
   Copyright (c) 2000 Steven Lawrance
   slawrance@technologist.com

   Released under the General Public License (GPL), which
   is available in the COPYING file. */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or 
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,     
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/version.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/msdos_fs.h>
#include <stdio.h>
#include <string.h>

void dfread(void *buffer, int size, int count, FILE *fd)
{
  /* Do sector-aligned reads */
  char sector[512];
  size *= count;
  while (size) {
    if (!fread(sector, 512, 1, fd)) {
      printf("Error reading file!\n");
      exit();
    }
    memcpy(buffer, sector, (size < 512)? size: 512);
    size -= (size < 512)? size: 512;
  }
}

/* Cluster-to-Byte */
#define C2B(boot, cluster) \
  (sizeof(struct msdos_dir_entry) * *((__u16 *) &boot.dir_entries[0]) + boot.fat_length * *((__u16 *) \
  &boot.sector_size[0]) * 2 + 512 + (cluster - 2) * boot.cluster_size * *((__u16 *) &boot.sector_size[0]))

/*#define IS_VALID_CLUSTER(boot, cluster) */

__u16 get_next_cluster(struct fat_boot_sector *boot, __u16 cluster, __u8 fat, FILE *f, __u8 fat_bits)
{
  /* Read the next cluster from the given FAT */
  __u16 next;
  fseek(f, (fat - 1) * (boot->fat_length * *((__u16 *) &(boot->sector_size[0]))) + 512 + (cluster * fat_bits) / 8, SEEK_SET);
  dfread(&next, 2, 1, f);
  if (fat_bits == 16)
    return next;
  if (fat_bits == 12) {
    if ((cluster * fat_bits) % 8)
      next = (next & 0xFFF0) >> 4;
    else
      next = next & 0x0FFF;
    return next;
  }
  return 0;
}

int main(int argc, char **argv)
{
  FILE *f;
  struct fat_boot_sector boot;
  __u16 *sector_size = (__u16 *) &boot.sector_size[0];
  __u16 *root_entries = (__u16 *) &boot.dir_entries[0];
  __u32 sectors;
  __u16 cluster_low, cluster_high;
  __u8 fat_bits = 16;
  __u8 cur_fat = 1;
  char menu, dummy;
  if (argc == 1) {
    printf("Usage: %s disk.image\n",argv[0]);
    exit();
  }
  f = fopen(argv[1], "r");
  if (!f) {
    printf("Error opening %s for read!\n", argv[1]);
    exit();
  }
  dfread(&boot, sizeof(boot), 1, f);
  if (!boot.sectors[0] && !boot.sectors[1]) sectors = boot.total_sect; else sectors = *((__u16 *) &boot.sectors[0]);
  cluster_low = 2;
  cluster_high = (sectors - (1 + boot.fat_length * boot.fats + (*root_entries * sizeof(struct msdos_dir_entry)) / *sector_size)) / boot.cluster_size + cluster_low - 1;
  if ((sectors / boot.cluster_size) * 2 > boot.fat_length * *sector_size) fat_bits = 12;

  /* Enter recovery loop */
  printf("\n");
  while (1) {
    printf("Program Options:\n");
    printf("----------------\n");
    printf("F  FAT statistics, based on the boot sector\n");
    if (fat_bits <= 16)
      printf("R  List root directory\n");
    printf("D  List a single directory cluster\n");
    printf("C  Translate a cluster into a byte\n");
    printf("B  Translate a byte into a cluster\n");
    printf("A  Change FAT (current = %u)\n", cur_fat);
    printf("X  Exit\n");
    printf("\n:");
    scanf("%c", &menu);
    scanf("%c", &dummy);
    printf("\n");
    if ((fat_bits > 16) && ((menu == 'R') || (menu == 'r')))
      menu = ' ';

    switch(menu) {

    case 'F' :
    case 'f' :
      {
        printf("FAT Information:\n");
        printf("----------------\n");
        printf("Logical sector size: %lu\n", *sector_size);
        printf("Number of sectors: %lu\n", sectors);
        printf("Cluster size: %lu sectors, %lu bytes\n", boot.cluster_size, boot.cluster_size * *sector_size);
        printf("Number of FATs: %u\n", boot.fats);
        printf("FAT length: %lu sectors, %lu bytes\n", boot.fat_length, boot.fat_length * *sector_size);
        printf("FAT 1 starts at byte %lu\n", 512);
        printf("FAT 2 starts at byte %lu\n", boot.fat_length * *sector_size + 512);
        printf("Wasted space at end of each FAT: %lu bytes\n", boot.fat_length * *sector_size - ((((sectors / boot.cluster_size) * fat_bits) / 8) / *sector_size + (((((sectors / boot.cluster_size) * fat_bits) / 8) % *sector_size)? 1: 0)) * *sector_size);
        printf("Root entries: %lu entries, starting at %lu\n", *root_entries, boot.fat_length * *sector_size * boot.fats + 512);
        printf("Number of clusters: %lu\n", sectors / boot.cluster_size);
        printf("Usable clusters: %u to %u\n", cluster_low, cluster_high);
        printf("Cluster %u begins at byte %lu\n", cluster_low, C2B(boot, cluster_low));
        printf("Cluster %u begins at byte %lu\n", cluster_high, C2B(boot, cluster_high));
        printf("Raw disk size: %lu bytes\n", sectors * *sector_size);
        printf("Maximum usable file space w/o directories: %lu bytes\n", (cluster_high - cluster_low + 1) * boot.cluster_size * *sector_size);
        printf("Wasted space at end of disk: %lu bytes\n", ((sectors / boot.cluster_size) - (cluster_high - cluster_low) - 1) * (boot.cluster_size * *sector_size) - C2B(boot, 2));
        break;
      }

    case 'r' :
    case 'R' :
    case 'D' :
    case 'd' :
      {
        /* List the root directory in FAT12 or FAT16 (in FAT32, / is a normal directory) */
        struct msdos_dir_entry cur;
        char lfn[256];
        __u16 cluster, prev;
        __u32 filepos;
        __u8 bad_chain;
        __u16 length = *root_entries;
        __u32 basepos = boot.fat_length * *sector_size * boot.fats + 512;
        int i, t = 0, lfns;
        lfn[247] = lfn[0] = 0;
        lfns = 0;

        /* Perform special processing if we have to ask for the byte */
        if ((menu == 'd') || (menu == 'D')) {
          printf("Enter the byte position of the cluster to examine: ");
          scanf("%lu", &basepos);
          scanf("%c", &dummy);
          length = (boot.cluster_size * *sector_size) / sizeof(struct msdos_dir_entry);
        }

        /* Read the directory entries */
        for (i = 0; i < length; i++) {
          bad_chain = 0;
/*          fseek(f, boot.fat_length * *sector_size * boot.fats + 512 + i * sizeof(cur), SEEK_SET);*/
          fseek(f, basepos + i * sizeof(cur), SEEK_SET); /* TEMP */
          dfread(&cur, sizeof(cur), 1, f);
          if (cur.name[0] && (cur.attr != ATTR_EXT) && (cur.attr != ATTR_VOLUME)) {
            printf("%8.8s.%3.3s (%lu bytes", cur.name, cur.ext, cur.size);
            if (lfns) {
              printf("; aka \"%s\") ", lfn);
              lfn[0] = 0;
              lfns = 0;
            } else
              printf(") ");

            /* If deleted, indicate it */
            if ((__u8) cur.name[0] == DELETED_FLAG)
              printf("DELETED ");

            /* Walk the FAT to determine where this is */
            cluster = cur.start;
            if (!cluster)
              printf("NULL START CLUSTER ");
            prev = 0;
            do {
              cur.size -= (cur.size > boot.cluster_size * *sector_size)? boot.cluster_size * *sector_size: cur.size;
              if (prev != cluster - 1) {
                if (!prev)
                  printf("%lu", C2B(boot, cluster));
                else
                  printf("-%lu %lu", C2B(boot, prev + 1) - 1, C2B(boot, cluster));
              }
              prev = cluster;
              cluster = get_next_cluster(&boot, cluster, cur_fat, f, fat_bits);
              if (((cluster >= 2) && (cluster < ((fat_bits == 16)? 0xFFF0: 0xFF0))) && !cur.size) {
                /* Infinite loop prevention and sanity check */
                bad_chain = 1;
                cluster = 0;
              }
            } while ((cluster >= 2) && (cluster < ((fat_bits == 16)? 0xFFF0: 0xFF0)));
            if (prev)
              printf("-%lu", C2B(boot, prev + 1) - 1);
            if (cur.size)
              printf(" CHAIN TOO SHORT");
            if (bad_chain)
              printf(" INVALID CHAIN");
            printf("\n");
            t++;
          } else if ((cur.attr == ATTR_EXT) && (lfns < 19)) /* max of 19 LFNs I think.. (247 chars max?) */
          {
            /* Grab this LFN */
            struct msdos_dir_slot *dirslot = (struct msdos_dir_slot *) &cur;
            int ch, pos;
            if (lfns)
              memmove(&(lfn[13]), lfn, 13 * lfns);
            else
              lfn[13] = 0;

            /* Very sloppy UNICODE hack ;-) */
            for (ch = 0, pos = 0; ch < 10; ch += 2, pos++)
              lfn[pos] = dirslot->name0_4[ch];
            for (ch = 0, pos = 5; ch < 12; ch += 2, pos++)
              lfn[pos] = dirslot->name5_10[ch];
            for (ch = 0, pos = 11; ch < 4; ch += 2, pos++)
              lfn[pos] = dirslot->name11_12[ch];
            lfns++;
          }

          if (t >= 10) {
            printf(":");
            scanf("%c", &dummy);
            printf("\n\n");
            t = 0;
          }
        }
        break;
      }

    case 'c' :
    case 'C' :
      {
        __u16 cluster;
        printf("Enter a cluster number: ");
        scanf("%u", &cluster);
        scanf("%c", &dummy);
        if ((cluster >= cluster_low) && (cluster <= cluster_high))
          printf("Cluster %u starts at byte %lu\n", cluster, C2B(boot, cluster));
        else
          printf("Invalid cluster number %u\n", cluster);
        break;
      }

    case 'b' :
    case 'B' :
      {
        __u32 byte;
        printf("Enter a byte position: ");
        scanf("%lu", &byte);
        scanf("%c", &dummy);
        if ((byte >= C2B(boot, cluster_low)) && (byte <= sectors * *sector_size))
          printf("Byte %lu exists in cluster %u, offset %u bytes\n", byte, (byte - C2B(boot, cluster_low)) / (boot.cluster_size * *sector_size) + cluster_low, (byte - C2B(boot, cluster_low)) % (boot.cluster_size * *sector_size));
        else
          printf("Byte %lu is not within a cluster (too low or too high)\n", byte);
        break;
      }

    case 'a' :
    case 'A' :
      {
        __u8 new_fat;
        printf("Enter a FAT from 1 to %u: ", boot.fats);
        scanf("%u", &new_fat);
        scanf("%c", &dummy);
        if ((new_fat >= 1) && (new_fat <= boot.fats))
          cur_fat = new_fat;
        else
          printf("Invalid FAT number %u\n", new_fat);
        break;
      }

    case 'x' :
    case 'X' :
      {
        printf("Thanks for using FAT Recover :-)\n");
        exit();
      }
    }

    /* Wait for the user to press a key */
    printf(":");
    scanf("%c", &dummy);
    printf("\n\n");
  }
  fclose(f);
  return 0;
}
